"""
Fat links related views
"""

# Standard Library
from datetime import timedelta

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.db import IntegrityError
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.datetime_safe import datetime
from django.utils.safestring import mark_safe
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.decorators import permissions_required
from allianceauth.eveonline.models import EveCharacter
from allianceauth.eveonline.providers import provider
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.forms import (
    AFatClickFatForm,
    AFatEsiFatForm,
    AFatManualFatForm,
    FatLinkEditForm,
)
from afat.helper.fatlinks import get_doctrines, get_esi_fleet_information_by_user
from afat.helper.time import get_time_delta
from afat.helper.views import convert_fatlinks_to_dict, convert_fats_to_dict
from afat.models import (
    Duration,
    Fat,
    FatLink,
    FleetType,
    Log,
    Setting,
    get_hash_on_save,
)
from afat.providers import esi
from afat.tasks import process_fats
from afat.utils import get_or_create_character, write_log

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required()
@permission_required("afat.basic_access")
def overview(request: WSGIRequest, year: int = None) -> HttpResponse:
    """
    Fat links view

    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    if year is None:
        year = datetime.now().year

    context = {
        "year": year,
        "year_current": datetime.now().year,
        "year_prev": int(year) - 1,
        "year_next": int(year) + 1,
    }

    logger.info(msg=f"FAT link list called by {request.user}")

    return render(
        request=request,
        template_name="afat/view/fatlinks/fatlinks-overview.html",
        context=context,
    )


@login_required()
@permission_required("afat.basic_access")
def ajax_get_fatlinks_by_year(request: WSGIRequest, year: int) -> JsonResponse:
    """
    Ajax call :: get all FAT links for a given year

    :param request:
    :type request:
    :param year:
    :type year:
    :return:
    :rtype:
    """

    fatlinks = (
        FatLink.objects.select_related_default()
        .filter(created__year=year)
        .annotate_fats_count()
    )

    fatlink_rows = [
        convert_fatlinks_to_dict(
            request=request,
            fatlink=fatlink,
            close_esi_redirect=reverse(viewname="afat:fatlinks_overview"),
        )
        for fatlink in fatlinks
    ]

    return JsonResponse(data=fatlink_rows, safe=False)


@login_required()
@permissions_required(("afat.manage_afat", "afat.add_fatlink"))
def add_fatlink(request: WSGIRequest) -> HttpResponse:
    """
    Add fat link view

    :param request:
    :type request:
    :return:
    :rtype:
    """

    fleet_types = FleetType.objects.filter(is_enabled=True).order_by("name")

    context = {
        "default_expiry_time": Setting.get_setting(
            Setting.Field.DEFAULT_FATLINK_EXPIRY_TIME
        ),
        "esi_fleet": get_esi_fleet_information_by_user(request.user),
        "esi_fatlink_form": AFatEsiFatForm(),
        "manual_fatlink_form": AFatClickFatForm(),
        "doctrines": get_doctrines(),
        "fleet_types": fleet_types,
    }

    logger.info(msg=f"Add FAT link view called by {request.user}")

    return render(
        request=request,
        template_name="afat/view/fatlinks/fatlinks-add-fatlink.html",
        context=context,
    )


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
def create_clickable_fatlink(
    request: WSGIRequest,
) -> HttpResponseRedirect:
    """
    Create a clickable fat link

    :param request:
    :type request:
    :return:
    :rtype:
    """

    if request.method == "POST":
        form = AFatClickFatForm(data=request.POST)

        if form.is_valid():
            fatlink_hash = get_hash_on_save()

            fatlink = FatLink()
            fatlink.fleet = form.cleaned_data["name"]
            fatlink.fleet_type = form.cleaned_data["type"]
            fatlink.doctrine = form.cleaned_data["doctrine"]
            fatlink.creator = request.user
            fatlink.hash = fatlink_hash
            fatlink.created = timezone.now()
            fatlink.save()

            dur = Duration()
            dur.fleet = FatLink.objects.get(hash=fatlink_hash)
            dur.duration = form.cleaned_data["duration"]
            dur.save()

            # Writing DB log
            fleet_type = (
                f" (Fleet type: {fatlink.fleet_type})" if fatlink.fleet_type else ""
            )

            write_log(
                request=request,
                log_event=Log.Event.CREATE_FATLINK,
                log_text=(
                    f'FAT link with name "{form.cleaned_data["name"]}"{fleet_type} and '
                    f'a duration of {form.cleaned_data["duration"]} minutes was created'
                ),
                fatlink_hash=fatlink.hash,
            )

            logger.info(
                msg=(
                    f'FAT link "{fatlink_hash}" with name '
                    f'"{form.cleaned_data["name"]}"{fleet_type} and a duration '
                    f'of {form.cleaned_data["duration"]} minutes was created '
                    f"by {request.user}"
                )
            )

            messages.success(
                request=request,
                message=mark_safe(
                    s=_(
                        "<h4>Success!</h4>"
                        "<p>Clickable FAT link created!</p>"
                        "<p>Make sure to give your fleet members the link to "
                        "click so that they get credit for this fleet.</p>"
                    )
                ),
            )

            return redirect(
                to="afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash
            )

        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4>"
                    "<p>Something went wrong when attempting "
                    "to submit your clickable FAT link.</p>"
                )
            ),
        )

        return redirect("afat:fatlinks_add_fatlink")

    messages.warning(
        request=request,
        message=mark_safe(
            s=_(
                "<h4>Warning!</h4>"
                '<p>You must fill out the form on the "Add FAT link" '
                "page to create a clickable FAT link</p>"
            )
        ),
    )

    return redirect(to="afat:fatlinks_add_fatlink")


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
@token_required(scopes=["esi-fleets.read_fleet.v1"])
def create_esi_fatlink_callback(  # pylint: disable=too-many-locals
    request: WSGIRequest, token, fatlink_hash: str
) -> HttpResponseRedirect:
    """
    Helper :: create ESI link (callback, used when coming back from character selection)

    :param request:
    :type request:
    :param token:
    :type token:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    # Check if there is a fleet
    try:
        required_scopes = ["esi-fleets.read_fleet.v1"]
        esi_token = Token.get_token(
            character_id=token.character_id, scopes=required_scopes
        )

        fleet_from_esi = esi.client.Fleets.get_characters_character_id_fleet(
            character_id=token.character_id, token=esi_token.valid_access_token()
        ).result()
    except Exception:  # pylint: disable=broad-exception-caught
        # Not in a fleet
        messages.warning(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Warning!</h4>"
                    "<p>To use the ESI function, you need to be in fleet and you need "
                    "to be the fleet boss! You can create a clickable FAT link and "
                    "share it, if you like.</p>"
                )
            ),
        )

        # Return to "Add FAT link" view
        return redirect(to="afat:fatlinks_add_fatlink")

    # check if this character already has a fleet
    creator_character = EveCharacter.objects.get(character_id=token.character_id)
    registered_fleets_for_creator = FatLink.objects.select_related_default().filter(
        is_esilink=True,
        is_registered_on_esi=True,
        character__character_name=creator_character.character_name,
    )

    fleet_already_registered = False
    character_has_registered_fleets = False
    registered_fleets_to_close = []

    if registered_fleets_for_creator.count() > 0:
        character_has_registered_fleets = True

        for registered_fleet in registered_fleets_for_creator:
            if registered_fleet.esi_fleet_id == fleet_from_esi["fleet_id"]:
                # Character already has a fleet
                fleet_already_registered = True
            else:
                registered_fleets_to_close.append(
                    {"registered_fleet": registered_fleet}
                )

    # If the FC already has a fleet, and it is the same as already registered,
    # just throw a warning
    if fleet_already_registered is True:
        messages.warning(
            request=request,
            message=mark_safe(
                s=format_lazy(
                    _(
                        "<h4>Warning!</h4>"
                        '<p>Fleet with ID "{fleet_id}" for your character '
                        "{creator__character_name} has already been registered and "
                        "pilots joining this fleet are automatically tracked.</p>"
                    ),
                    fleet_id=fleet_from_esi["fleet_id"],
                    creator__character_name=creator_character.character_name,
                ),
            ),
        )

        # Return to "Add FAT link" view
        return redirect(to="afat:fatlinks_add_fatlink")

    # If it's a new fleet, remove all former registered fleets, if there are any
    if (
        character_has_registered_fleets is True
        and fleet_already_registered is False
        and len(registered_fleets_to_close) > 0
    ):
        for registered_fleet_to_close in registered_fleets_to_close:
            reason = (
                f"FC has opened a new fleet with the "
                f"character {creator_character.character_name}"
            )

            logger.info(
                msg=(
                    "Closing ESI FAT link with hash "
                    f'"{registered_fleet_to_close["registered_fleet"].hash}". '
                    f"Reason: {reason}"
                )
            )

            registered_fleet_to_close["registered_fleet"].is_registered_on_esi = False
            registered_fleet_to_close["registered_fleet"].save()

    # Check if we deal with the fleet boss here
    try:
        esi_fleet_member = esi.client.Fleets.get_fleets_fleet_id_members(
            fleet_id=fleet_from_esi["fleet_id"],
            token=esi_token.valid_access_token(),
        ).result()
    except Exception:  # pylint: disable=broad-exception-caught
        messages.warning(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Warning!</h4>"
                    "<p>Not Fleet Boss! Only the fleet boss can utilize the ESI "
                    "function. You can create a clickable FAT link and share it, "
                    "if you like.</p>"
                )
            ),
        )

        # Return to "Add FAT link" view
        return redirect(to="afat:fatlinks_add_fatlink")

    creator_character = EveCharacter.objects.get(character_id=token.character_id)

    # Create the fat link
    fatlink = FatLink(
        created=timezone.now(),
        fleet=request.session["fatlink_form__name"],
        doctrine=request.session["fatlink_form__doctrine"],
        creator=request.user,
        character=creator_character,
        hash=fatlink_hash,
        is_esilink=True,
        is_registered_on_esi=True,
        esi_fleet_id=fleet_from_esi["fleet_id"],
        fleet_type=request.session["fatlink_form__type"],
    )

    # Save it
    fatlink.save()

    # Writing DB log
    fleet_type = f"(Fleet type: {fatlink.fleet_type})" if fatlink.fleet_type else ""

    write_log(
        request=request,
        log_event=Log.Event.CREATE_FATLINK,
        log_text=(
            f'ESI FAT link with name "{request.session["fatlink_form__name"]}" '
            f"{fleet_type} was created by {request.user}"
        ),
        fatlink_hash=fatlink.hash,
    )

    logger.info(
        msg=(
            f'ESI FAT link "{fatlink_hash}" with name '
            f'"{request.session["fatlink_form__name"]}" {fleet_type} '
            f"was created by {request.user}"
        )
    )

    # Clear session
    del request.session["fatlink_form__name"]
    del request.session["fatlink_form__type"]

    # Process fleet members in the background
    process_fats.delay(
        data_list=esi_fleet_member, data_source="esi", fatlink_hash=fatlink_hash
    )

    messages.success(
        request=request,
        message=mark_safe(
            s=_(
                "<h4>Success!</h4>"
                "<p>FAT link Created!</p>"
                "<p>FATs have been queued, they may take a few minutes to show up.</p>"
                "<p>Pilots who join later will be automatically added until you "
                "close or leave the fleet in-game.</p>"
            )
        ),
    )

    return redirect(to="afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)


@login_required()
def create_esi_fatlink(
    request: WSGIRequest,
) -> HttpResponseRedirect:
    """
    Create an ESI fat link

    :param request:
    :type request:
    :return:
    :rtype:
    """

    fatlink_form = AFatEsiFatForm(data=request.POST)

    if fatlink_form.is_valid():
        fatlink_hash = get_hash_on_save()

        request.session["fatlink_form__name"] = fatlink_form.cleaned_data["name_esi"]
        request.session["fatlink_form__doctrine"] = fatlink_form.cleaned_data[
            "doctrine_esi"
        ]
        request.session["fatlink_form__type"] = fatlink_form.cleaned_data["type_esi"]

        return redirect(
            to="afat:fatlinks_create_esi_fatlink_callback", fatlink_hash=fatlink_hash
        )

    messages.error(
        request=request,
        message=mark_safe(
            s=_(
                "<h4>Error!</h4>"
                "<p>Something went wrong when attempting to "
                "submit your ESI FAT link.</p>"
            )
        ),
    )

    return redirect(to="afat:fatlinks_add_fatlink")


@login_required()
@permission_required(perm="afat.basic_access")
@token_required(
    scopes=[
        "esi-location.read_location.v1",
        "esi-location.read_ship_type.v1",
        "esi-location.read_online.v1",
    ]
)
def add_fat(
    request: WSGIRequest, token, fatlink_hash: str = None
) -> HttpResponseRedirect:
    """
    Click fat link helper
    :param request:
    :type request:
    :param token:
    :type token:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    if fatlink_hash is None:
        messages.warning(
            request=request,
            message=mark_safe(
                s=_("<h4>Warning!</h4><p>No FAT link hash provided.</p>")
            ),
        )

        return redirect(to="afat:dashboard")

    try:
        fleet = FatLink.objects.get(hash=fatlink_hash, is_esilink=False)
    except FatLink.DoesNotExist:
        messages.warning(
            request=request,
            message=mark_safe(
                s=_("<h4>Warning!</h4><p>The hash provided is not valid.</p>")
            ),
        )

        return redirect(to="afat:dashboard")

    dur = Duration.objects.get(fleet=fleet)
    now = timezone.now() - timedelta(minutes=dur.duration)

    if now >= fleet.created:
        messages.warning(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Warning!</h4>"
                    "<p>Sorry, that FAT link is expired. "
                    "If you were on that fleet, contact your FC about "
                    "having your FAT manually added.</p>"
                )
            ),
        )

        return redirect(to="afat:dashboard")

    character = EveCharacter.objects.get(character_id=token.character_id)

    try:
        required_scopes = [
            "esi-location.read_location.v1",
            "esi-location.read_online.v1",
            "esi-location.read_ship_type.v1",
        ]
        esi_token = Token.get_token(
            character_id=token.character_id, scopes=required_scopes
        )
    except Exception:  # pylint: disable=broad-exception-caught
        messages.warning(
            request=request,
            message=mark_safe(
                s=format_lazy(
                    _(
                        "<h4>Warning!</h4>"
                        "<p>There was an issue with the ESI token for {character_name}. "
                        "Please try again.</p>"
                    ),
                    character_name=character.character_name,
                )
            ),
        )

        return redirect(to="afat:dashboard")

    # Check if character is online
    character_online = esi.client.Location.get_characters_character_id_online(
        character_id=token.character_id, token=esi_token.valid_access_token()
    ).result()

    if character_online["online"] is True:
        # Character location
        location = esi.client.Location.get_characters_character_id_location(
            character_id=token.character_id,
            token=esi_token.valid_access_token(),
        ).result()

        # Current ship
        ship = esi.client.Location.get_characters_character_id_ship(
            character_id=token.character_id,
            token=esi_token.valid_access_token(),
        ).result()

        # System information
        system = esi.client.Universe.get_universe_systems_system_id(
            system_id=location["solar_system_id"]
        ).result()["name"]

        ship_name = provider.get_itemtype(type_id=ship["ship_type_id"]).name

        try:
            Fat(
                fatlink=fleet,
                character=character,
                system=system,
                shiptype=ship_name,
                corporation_eve_id=character.corporation_id,
                alliance_eve_id=character.alliance_id,
            ).save()
        except IntegrityError:
            messages.warning(
                request=request,
                message=mark_safe(
                    s=format_lazy(
                        _(
                            "<h4>Warning!</h4>"
                            "<p>The selected charcter ({character_name}) is already "
                            "registered for this FAT link.</p>"
                        ),
                        character_name=character.character_name,
                    )
                ),
            )
        else:
            if fleet.fleet is not None:
                fleet_name = fleet.fleet
            else:
                fleet_name = fleet.hash

            messages.success(
                request=request,
                message=mark_safe(
                    s=format_lazy(
                        _(
                            "<h4>Success!</h4>"
                            '<p>FAT registered for {character_name} at "{fleet_name}"</p>'
                        ),
                        character_name=character.character_name,
                        fleet_name=fleet_name,
                    )
                ),
            )

            logger.info(
                msg=(
                    f'Participation for fleet "{fleet_name}" registered for '
                    f"pilot {character.character_name}"
                )
            )
    else:
        messages.warning(
            request=request,
            message=mark_safe(
                s=format_lazy(
                    _(
                        "<h4>Warning!</h4>"
                        "<p>Cannot register the fleet participation for "
                        "{character_name}. The character needs to be online.</p>"
                    ),
                    character_name=character.character_name,
                )
            ),
        )

    return redirect(to="afat:dashboard")


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
def details_fatlink(  # pylint: disable=too-many-statements too-many-branches too-many-locals
    request: WSGIRequest, fatlink_hash: str
) -> HttpResponse:
    """
    Fat link view

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        link = FatLink.objects.select_related_default().get(hash=fatlink_hash)
    except FatLink.DoesNotExist:
        messages.warning(
            request=request,
            message=mark_safe(
                s=_("<h4>Warning!</h4><p>The hash provided is not valid.</p>")
            ),
        )

        return redirect(to="afat:dashboard")

    if request.method == "POST":
        fatlink_edit_form = FatLinkEditForm(data=request.POST)
        manual_fat_form = AFatManualFatForm(data=request.POST)

        if fatlink_edit_form.is_valid():
            link.fleet = fatlink_edit_form.cleaned_data["fleet"]
            link.save()

            # Writing DB log
            write_log(
                request=request,
                log_event=Log.Event.CHANGE_FATLINK,
                log_text=f'FAT link changed. Fleet name was set to "{link.fleet}"',
                fatlink_hash=link.hash,
            )

            logger.info(
                msg=(
                    f'FAT link with hash "{link.hash}" changed. '
                    f'Fleet name was set to "{link.fleet}" by {request.user}'
                )
            )

            messages.success(
                request=request,
                message=mark_safe(
                    s=_("<h4>Success!</h4><p>Fleet name successfully changed.</p>")
                ),
            )
        elif manual_fat_form.is_valid():
            character_name = manual_fat_form.cleaned_data["character"]
            system = manual_fat_form.cleaned_data["system"]
            shiptype = manual_fat_form.cleaned_data["shiptype"]
            character = get_or_create_character(name=character_name)

            if character is not None:
                afat_object, created = Fat.objects.get_or_create(
                    fatlink=link,
                    character=character,
                    defaults={
                        "system": system,
                        "shiptype": shiptype,
                    },
                )

                if created is True:
                    messages.success(
                        request=request,
                        message=mark_safe(
                            s=format_lazy(
                                _(
                                    "<h4>Success!</h4><p>Manual FAT processed.<br>"
                                    "{character_name} has been added flying a {shiptype} "
                                    "in {system}</p>"
                                ),
                                character_name=character.character_name,
                                shiptype=shiptype,
                                system=system,
                            )
                        ),
                    )

                    # Writing DB log
                    write_log(
                        request=request,
                        log_event=Log.Event.MANUAL_FAT,
                        log_text=(
                            f"Pilot {character.character_name} "
                            f"flying a {shiptype} was manually added"
                        ),
                        fatlink_hash=link.hash,
                    )

                    logger.info(
                        msg=(
                            f"Pilot {character.character_name} flying a {shiptype} was "
                            "manually added to FAT link with "
                            f'hash "{link.hash}" by {request.user}'
                        )
                    )
                else:
                    messages.info(
                        request=request,
                        message=mark_safe(
                            s=format_lazy(
                                _(
                                    "<h4>Information</h4>"
                                    "<p>Pilot is already registered for this FAT link.</p>"
                                    "<p>Name: {character_name}<br>System: {system}<br>Ship: {shiptype}</p>"
                                ),
                                character_name=character.character_name,
                                system=afat_object.system,
                                shiptype=afat_object.shiptype,
                            )
                        ),
                    )
            else:
                messages.error(
                    request=request,
                    message=mark_safe(
                        s=_(
                            "<h4>Oh No!</h4>"
                            "<p>Manual FAT processing failed! "
                            "The character name you entered was not found.</p>"
                        )
                    ),
                )
        else:
            messages.error(
                request=request,
                message=mark_safe(s=_("<h4>Oh No!</h4><p>Something went wrong!</p>")),
            )

    logger.info(msg=f'FAT link "{fatlink_hash}" details view called by {request.user}')

    # Let's see if the link is still valid or has expired already and can be re-opened
    # and FATs can be manually added
    # (only possible for 24 hours after creating the FAT link)
    link_ongoing = True
    link_can_be_reopened = False
    link_expires = None
    manual_fat_can_be_added = False

    # Time dependant settings
    try:
        dur = Duration.objects.get(fleet=link)
    except Duration.DoesNotExist:
        # ESI link
        link_ongoing = False
    else:
        link_expires = link.created + timedelta(minutes=dur.duration)
        now = timezone.now()

        if link_expires <= now:
            # Link expired
            link_ongoing = False

            if link.reopened is False and get_time_delta(
                then=link_expires, now=now, interval="minutes"
            ) < Setting.get_setting(Setting.Field.DEFAULT_FATLINK_REOPEN_GRACE_TIME):
                link_can_be_reopened = True

        # Manual fat still possible?
        # Only possible if the FAT link has not been re-opened
        # and has been created within the last 24 hours
        if (
            link.reopened is False
            and get_time_delta(then=link.created, now=now, interval="hours") < 24
        ):
            manual_fat_can_be_added = True

    is_clickable_link = False
    if link.is_esilink is False:
        is_clickable_link = True

    if link.is_esilink and link.is_registered_on_esi:
        link_ongoing = True

    context = {
        "link": link,
        "is_esi_link": link.is_esilink,
        "is_clickable_link": is_clickable_link,
        "link_expires": link_expires,
        "link_ongoing": link_ongoing,
        "link_can_be_reopened": link_can_be_reopened,
        "manual_fat_can_be_added": manual_fat_can_be_added,
        "reopen_grace_time": Setting.get_setting(
            Setting.Field.DEFAULT_FATLINK_REOPEN_GRACE_TIME
        ),
        "reopen_duration": Setting.get_setting(
            Setting.Field.DEFAULT_FATLINK_REOPEN_DURATION
        ),
    }

    return render(
        request=request,
        template_name="afat/view/fatlinks/fatlinks-details-fatlink.html",
        context=context,
    )


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
def ajax_get_fats_by_fatlink(request: WSGIRequest, fatlink_hash) -> JsonResponse:
    """
    Ajax call :: get all FATs for a given FAT link hash

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    fats = Fat.objects.select_related_default().filter(fatlink__hash=fatlink_hash)

    fat_rows = [convert_fats_to_dict(request=request, fat=fat) for fat in fats]

    return JsonResponse(data=fat_rows, safe=False)


@login_required()
@permission_required(perm="afat.manage_afat")
def delete_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponseRedirect:
    """
    Delete fat link helper

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        link = FatLink.objects.get(hash=fatlink_hash)
    except FatLink.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>The FAT link hash provided is either invalid "
                    "or the FAT link has already been deleted.</p>"
                )
            ),
        )

        return redirect("afat:dashboard")

    # Delete associated FATs and the FAT link
    Fat.objects.filter(fatlink_id=link.pk).delete()
    link.delete()

    # Log the deletion
    write_log(
        request,
        log_event=Log.Event.DELETE_FATLINK,
        log_text="FAT link deleted.",
        fatlink_hash=fatlink_hash,
    )

    messages.success(
        request,
        mark_safe(
            format_lazy(
                _(
                    "<h4>Success!</h4>"
                    '<p>The FAT link "{fatlink_hash}" and all associated FATs have '
                    "been successfully deleted.</p>"
                ),
                fatlink_hash=fatlink_hash,
            )
        ),
    )

    logger.info(
        f'Fat link "{fatlink_hash}" and all associated FATs have been deleted by {request.user}'
    )

    return redirect("afat:fatlinks_overview")


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.delete_afat"))
def delete_fat(
    request: WSGIRequest, fatlink_hash: str, fat_id: int
) -> HttpResponseRedirect:
    """
    Delete fat helper

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :param fat_id:
    :type fat_id:
    :return:
    :rtype:
    """

    try:
        link = FatLink.objects.get(hash=fatlink_hash)
        fat_details = Fat.objects.get(pk=fat_id, fatlink_id=link.pk)
    except FatLink.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4>"
                    "<p>The hash provided is either invalid or has been deleted.</p>"
                )
            ),
        )

        return redirect(to="afat:dashboard")
    except Fat.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>The hash and FAT ID do not match.</p>")
            ),
        )

        return redirect(to="afat:dashboard")

    fat_details.delete()

    write_log(
        request=request,
        log_event=Log.Event.DELETE_FAT,
        log_text=f"The FAT for {fat_details.character.character_name} has been deleted",
        fatlink_hash=link.hash,
    )

    messages.success(
        request=request,
        message=mark_safe(
            s=format_lazy(
                _(
                    "<h4>Success!</h4>"
                    "<p>The FAT for {character_name} has been successfully deleted "
                    'from FAT link "{fatlink_hash}".</p>'
                ),
                character_name=fat_details.character.character_name,
                fatlink_hash=fatlink_hash,
            )
        ),
    )

    logger.info(
        msg=(
            f"The FAT for {fat_details.character.character_name} has "
            f'been deleted from FAT link "{fatlink_hash}" by {request.user}.'
        )
    )

    return redirect(to="afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
def close_esi_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponseRedirect:
    """
    Ajax call to close an ESI fat link

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        fatlink = FatLink.objects.get(hash=fatlink_hash)
        fatlink.is_registered_on_esi = False
        fatlink.save()

        logger.info(
            msg=(
                f'Closing ESI FAT link with hash "{fatlink_hash}". '
                "Reason: Closed by manual request"
            )
        )
    except FatLink.DoesNotExist:
        logger.info(msg=f'ESI FAT link with hash "{fatlink_hash}" does not exist')

    next_view = request.GET.get("next", reverse("afat:dashboard"))

    return HttpResponseRedirect(redirect_to=next_view)


@login_required()
@permissions_required(perm=("afat.manage_afat", "afat.add_fatlink"))
def reopen_fatlink(request: WSGIRequest, fatlink_hash: str) -> HttpResponseRedirect:
    """
    Re-open fat link

    :param request:
    :type request:
    :param fatlink_hash:
    :type fatlink_hash:
    :return:
    :rtype:
    """

    try:
        fatlink_duration = Duration.objects.get(fleet__hash=fatlink_hash)
    except Duration.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4>"
                    "<p>The hash you provided does not match with any FAT link.</p>"
                )
            ),
        )

        return redirect(to="afat:dashboard")

    if not fatlink_duration.fleet.reopened:
        created_at = fatlink_duration.fleet.created
        now = datetime.now()

        default_reopen_duration = Setting.get_setting(
            Setting.Field.DEFAULT_FATLINK_REOPEN_DURATION
        )
        time_difference_in_minutes = get_time_delta(
            then=created_at, now=now, interval="minutes"
        )
        new_duration = time_difference_in_minutes + default_reopen_duration

        fatlink_duration.duration = new_duration
        fatlink_duration.save()

        fatlink_duration.fleet.reopened = True
        fatlink_duration.fleet.save()

        # writing DB log
        write_log(
            request=request,
            log_event=Log.Event.REOPEN_FATLINK,
            log_text=f"FAT link re-opened for a duration of {default_reopen_duration} minutes",
            fatlink_hash=fatlink_duration.fleet.hash,
        )

        logger.info(
            msg=(
                f'FAT link with hash "{fatlink_hash}" re-opened by {request.user} '
                f"for a duration of {default_reopen_duration} minutes"
            )
        )

        messages.success(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Success!</h4>"
                    "<p>The FAT link has been successfully re-opened.</p>"
                )
            ),
        )
    else:
        messages.warning(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Warning!</h4>"
                    "<p>This FAT link has already been re-opened. "
                    "FAT links can be re-opened only once!</p>"
                )
            ),
        )

    return redirect(to="afat:fatlinks_details_fatlink", fatlink_hash=fatlink_hash)



"""
URL patterns for the logs
These URLs are prefixed with `logs/`
"""

# Django
from django.urls import path

# Alliance Auth AFAT
from afat.views import logs

urls = [path(route="", view=logs.overview, name="logs_overview")]



"""
Test our template tags
"""

# Django
from django.template import Context, Template

# Alliance Auth AFAT
from afat.tests import BaseTestCase


class TestAfatFilters(BaseTestCase):
    """
    Test template filters
    """

    def test_month_name_filter(self):
        """
        Test month_name

        :return:
        """

        context = Context(dict_={"month": 5})
        template_to_render = Template(
            template_string="{% load afat %} {{ month|month_name }}"
        )

        rendered_template = template_to_render.render(context=context)

        self.assertInHTML(needle="May", haystack=rendered_template)


class TestSumValuesFilter(BaseTestCase):
    """
    Test the sum_values filter
    """

    def test_sum_values(self):
        """
        Test sum_values

        :return:
        :rtype:
        """

        context = Context(dict_={"test_dict": {"a": 1, "b": 2, "c": 3}})
        template_to_render = Template(
            template_string="{% load afat %} {{ test_dict|sum_values }}"
        )

        rendered_template = template_to_render.render(context=context)

        self.assertInHTML(needle="6", haystack=rendered_template)



"""
Auth hooks
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# Alliance Auth AFAT
from afat import __title__, urls
from afat.app_settings import securegroups_installed
from afat.constants import APP_BASE_URL
from afat.models import FatsInTimeFilter


class AaAfatMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    This class ensures only authorized users will see the menu entry
    """

    def __init__(self):
        # Setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=__title__,
            classes="fa-solid fa-space-shuttle",
            url_name="afat:dashboard",
            navactive=["afat:"],
        )

    def render(self, request):
        """
        Only if the user has access to this app

        :param request:
        :type request:
        :return:
        :rtype:
        """

        if request.user.has_perm(perm="afat.basic_access"):
            return MenuItemHook.render(self, request=request)

        return ""


@hooks.register(name="menu_item_hook")
def register_menu():
    """
    Register our menu

    :return:
    :rtype:
    """

    return AaAfatMenuItem()


@hooks.register(name="url_hook")
def register_url():
    """
    Register our menu link

    :return:
    :rtype:
    """

    return UrlHook(urls=urls, namespace="afat", base_url=rf"^{APP_BASE_URL}/")


if securegroups_installed():

    @hooks.register("secure_group_filters")
    def filters() -> list:
        """
        Secure group filter

        :return: Secure group filters
        :rtype: list
        """

        return [FatsInTimeFilter]



# Standard Library
from http import HTTPStatus
from unittest.mock import Mock

# Django
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

# Alliance Auth AFAT
from afat.models import Fat, FatLink
from afat.tests import BaseTestCase
from afat.tests.fixtures.load_allianceauth import load_allianceauth
from afat.views.dashboard import overview

MODULE_PATH = "afat.views.dashboard"


def response_content_to_str(response) -> str:
    return response.content.decode(response.charset)


class TestDashboard(BaseTestCase):
    """
    Test the dashboard view
    """

    @classmethod
    def setUpClass(cls):
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()

        cls.factory = RequestFactory()

        load_allianceauth()

        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id,
            permissions=["afat.basic_access"],
        )

        add_character_to_user(user=cls.user, character=cls.character_1101)

        create_user_from_evecharacter(character_id=cls.character_1002.character_id)

        cls.afat_link = FatLink.objects.create(
            fleet="Demo Fleet",
            hash="123",
            creator=cls.user,
            character=cls.character_1001,
        )

    def _page_overview_request(self, user):
        """
        Make a request to the overview page as the given user.

        :param user:
        :type user:
        :return:
        :rtype:
        """

        request = self.factory.get(path=reverse(viewname="afat:dashboard"))
        request.user = user

        middleware = SessionMiddleware(get_response=Mock())
        middleware.process_request(request=request)

        return overview(request)

    def test_should_only_show_my_chars_and_only_those_with_fat_links(self):
        """
        Test that the overview page only shows the characters of the user that have FAT links.

        :return:
        :rtype:
        """

        Fat.objects.create(character=self.character_1101, fatlink=self.afat_link)
        Fat.objects.create(character=self.character_1002, fatlink=self.afat_link)

        response = self._page_overview_request(user=self.user)

        content = response_content_to_str(response=response)

        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)
        self.assertIn(
            member=f'<span class="d-block" id="afat-eve-character-id-{self.character_1101.character_id}">{self.character_1101.character_name}</span>',
            container=content,
        )
        self.assertNotIn(
            member=f'<span class="d-block" id="afat-eve-character-id-{self.character_1001.character_id}">{self.character_1001.character_name}</span>',
            container=content,
        )
        self.assertNotIn(
            member=f'<span class="d-block" id="afat-eve-character-id-{self.character_1002.character_id}">{self.character_1002.character_name}</span>',
            container=content,
        )

    def test_ajax_recent_get_fats_by_character_should_only_show_my_fatlinks(self):
        """
        Test that the ajax call only returns the FATs for the given character if the user has access to the character.

        :return:
        :rtype:
        """

        Fat.objects.create(character=self.character_1101, fatlink=self.afat_link)
        Fat.objects.create(character=self.character_1002, fatlink=self.afat_link)

        self.client.force_login(user=self.user)

        response_correct_char = self.client.get(
            path=reverse(
                viewname="afat:dashboard_ajax_get_recent_fats_by_character",
                kwargs={"charid": self.character_1101.character_id},
            )
        )

        response_wrong_char = self.client.get(
            path=reverse(
                viewname="afat:dashboard_ajax_get_recent_fats_by_character",
                kwargs={"charid": self.character_1002.character_id},
            )
        )

        self.assertEqual(first=response_correct_char.status_code, second=HTTPStatus.OK)
        self.assertNotEqual(first=response_correct_char.json(), second=[])

        self.assertEqual(first=response_wrong_char.status_code, second=HTTPStatus.OK)
        self.assertEqual(first=response_wrong_char.json(), second=[])



"""
The forms we use
"""

# Django
from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth AFAT
from afat.models import Doctrine, Setting


def get_mandatory_form_label_text(text):
    """
    Label text for mandatory form fields

    :param text:
    :type text:
    :return:
    :rtype:
    """

    required_text = _("This field is mandatory")
    required_marker = (
        f'<span aria-label="{required_text}" class="form-required-marker">*</span>'
    )

    return mark_safe(
        f'<span class="form-field-required">{text} {required_marker}</span>'
    )


class AFatEsiFatForm(forms.Form):
    """
    Fat link form
    Used to create ESI FAT links
    """

    name_esi = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Fleet name")),
        max_length=255,
        widget=forms.TextInput(attrs={"placeholder": _("Enter fleet name")}),
    )
    type_esi = forms.CharField(
        required=False,
        label=_("Fleet type (optional)"),
        widget=forms.TextInput(
            attrs={
                "data-datalist": "afat-fleet-type-list",
                "data-full-width": "true",
            }
        ),
    )
    doctrine_esi = forms.CharField(
        required=False,
        label=_("Doctrine (optional)"),
        widget=forms.TextInput(
            attrs={
                "data-datalist": "afat-fleet-doctrine-list",
                "data-full-width": "true",
            }
        ),
    )


class AFatManualFatForm(forms.Form):
    """
    Manual FAT form
    """

    character = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Character Name")),
        max_length=255,
    )
    system = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("System")),
        max_length=100,
    )
    shiptype = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Ship type")),
        max_length=100,
    )


class AFatClickFatForm(forms.Form):
    """
    Fat link form
    Used to create clickable FAT links
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Fleet name")),
        max_length=255,
        widget=forms.TextInput(attrs={"placeholder": _("Enter fleet name")}),
    )
    type = forms.CharField(
        required=False,
        label=_("Fleet type (optional)"),
        widget=forms.TextInput(
            attrs={
                "data-datalist": "afat-fleet-type-list",
                "data-full-width": "true",
            }
        ),
    )
    doctrine = forms.CharField(
        required=False,
        label=_("Doctrine (optional)"),
        widget=forms.TextInput(
            attrs={
                "data-datalist": "afat-fleet-doctrine-list",
                "data-full-width": "true",
            }
        ),
    )
    duration = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(text=_("FAT link expiry time in minutes")),
        min_value=1,
        initial=0,
        widget=forms.TextInput(attrs={"placeholder": _("Expiry time in minutes")}),
    )

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)

        # This is a hack to set the initial value of the duration field,
        # which comes from the settings in the database.
        self.fields["duration"].initial = Setting.get_setting(
            Setting.Field.DEFAULT_FATLINK_EXPIRY_TIME
        )


class FatLinkEditForm(forms.Form):
    """
    Fat link edit form
    Used in edit view to change the fat link name
    """

    fleet = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Fleet name")),
        max_length=255,
    )


class SettingAdminForm(forms.ModelForm):
    """
    Form definitions for the FleetType form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta
        """

        model = Setting

        fields = "__all__"


class DoctrineAdminForm(forms.ModelForm):
    """
    Form definitions for the Doctrine form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta
        """

        model = Doctrine

        fields = "__all__"



# Generated by Django 4.0.10 on 2023-09-22 11:04

# Django
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

# Alliance Auth AFAT
import afat.models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("afat", "0022_afatlink_esi_error_count"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="AFat",
            new_name="Fat",
        ),
        migrations.RenameField(
            model_name="fat",
            old_name="afatlink",
            new_name="fatlink",
        ),
        migrations.AlterField(
            model_name="fat",
            name="fatlink",
            field=models.ForeignKey(
                help_text="The FAT link the character registered at",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afat_fats",
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="fat",
            name="character",
            field=models.ForeignKey(
                help_text="Character who registered this FAT",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afat_fats",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.RenameModel(
            old_name="AFatLinkType",
            new_name="FleetType",
        ),
        migrations.AlterModelOptions(
            name="fleettype",
            options={
                "default_permissions": (),
                "verbose_name": "Fleet type",
                "verbose_name_plural": "Fleet types",
            },
        ),
        migrations.AlterField(
            model_name="fleettype",
            name="name",
            field=models.CharField(
                help_text="Descriptive name of the fleet type", max_length=254
            ),
        ),
        migrations.RenameModel(
            old_name="AFatLog",
            new_name="Log",
        ),
        migrations.AlterField(
            model_name="log",
            name="user",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=models.SET(afat.models.get_sentinel_user),
                related_name="afat_log",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterModelOptions(
            name="log",
            options={
                "default_permissions": (),
                "verbose_name": "Log",
                "verbose_name_plural": "Logs",
            },
        ),
        migrations.AlterField(
            model_name="log",
            name="log_event",
            field=models.CharField(
                choices=[
                    ("CR_FAT_LINK", "FAT link created"),
                    ("CH_FAT_LINK", "FAT link changed"),
                    ("RM_FAT_LINK", "FAT link removed"),
                    ("RO_FAT_LINK", "FAT link re-opened"),
                    ("RM_FAT", "FAT removed"),
                    ("CR_FAT_MAN", "Manual FAT added"),
                ],
                default="CR_FAT_LINK",
                max_length=11,
            ),
        ),
        migrations.RenameModel(
            old_name="AaAfat",
            new_name="General",
        ),
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access the AFAT module"),
                    ("manage_afat", "Can manage the AFAT module"),
                    ("add_fatlink", "Can create FAT links"),
                    ("stats_corporation_own", "Can see own corporation statistics"),
                    (
                        "stats_corporation_other",
                        "Can see statistics of other corporations",
                    ),
                    ("log_view", "Can view the modules log"),
                ),
                "verbose_name": "AFAT",
            },
        ),
        migrations.RenameModel(
            old_name="ClickAFatDuration",
            new_name="Duration",
        ),
        migrations.AlterModelOptions(
            name="duration",
            options={
                "default_permissions": (),
                "verbose_name": "FAT link duration",
                "verbose_name_plural": "FAT link durations",
            },
        ),
        migrations.AlterField(
            model_name="duration",
            name="fleet",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="duration",
                to="afat.afatlink",
            ),
        ),
        migrations.DeleteModel(
            name="ManualAFat",
        ),
        migrations.RenameModel(
            old_name="AFatLink",
            new_name="FatLink",
        ),
        migrations.AlterUniqueTogether(
            name="fat",
            unique_together={("character", "fatlink")},
        ),
        migrations.AlterModelOptions(
            name="fatlink",
            options={
                "default_permissions": (),
                "ordering": ("-created",),
                "verbose_name": "FAT link",
                "verbose_name_plural": "FAT links",
            },
        ),
        migrations.RenameField(
            model_name="fatlink",
            old_name="afattime",
            new_name="created",
        ),
    ]



"""
The models
"""

# Standard Library
from typing import Any, ClassVar

# Third Party
from solo.models import SingletonModel

# Django
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth AFAT
from afat.managers import FatLinkManager, FatManager


def get_sentinel_user() -> User:
    """
    Get user or create the sentinel user

    :return:
    """

    return User.objects.get_or_create(username="deleted")[0]


def get_hash_on_save() -> str:
    """
    Get the hash

    :return:
    """

    fatlink_hash = get_random_string(length=30)

    while FatLink.objects.filter(hash=fatlink_hash).exists():
        fatlink_hash = get_random_string(length=30)

    return fatlink_hash


class General(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        AaAfat :: Meta
        """

        managed = False
        default_permissions = ()
        permissions = (
            # can access and register his own participation to a FAT link
            ("basic_access", _("Can access the AFAT module")),
            # Can manage the FAT module
            # Has:
            #   » add_fatlink
            #   » change_fatlink
            #   » delete_fatlink
            #   » add_fat
            #   » delete_fat
            ("manage_afat", _("Can manage the AFAT module")),
            # Can add a new FAT link
            ("add_fatlink", _("Can create FAT links")),
            # Can see own corp stats
            ("stats_corporation_own", _("Can see own corporation statistics")),
            # Can see the stats of all corps
            ("stats_corporation_other", _("Can see statistics of other corporations")),
            # Can view the modules log
            ("log_view", _("Can view the modules log")),
        )
        verbose_name = _("AFAT")


class FleetType(models.Model):
    """
    FAT link fleet type

    Example:
        - CTA
        - Home Defense
        - StratOP
        - and so on …
    """

    id = models.AutoField(primary_key=True)

    name = models.CharField(
        max_length=254, help_text=_("Descriptive name of the fleet type")
    )

    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=_("Whether this fleet type is active or not"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Fleet type")
        verbose_name_plural = _("Fleet types")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return str(self.name)


class FatLink(models.Model):
    """
    FAT link
    """

    class EsiError(models.TextChoices):
        """
        Choices for SRP Status
        """

        NOT_IN_FLEET = "NOT_IN_FLEET", _(
            "FC is not in the registered fleet anymore or fleet is no longer available."
        )
        NO_FLEET = "NO_FLEET", _("Registered fleet seems to be no longer available.")
        NOT_FLEETBOSS = "NOT_FLEETBOSS", _("FC is no longer the fleet boss.")

    created = models.DateTimeField(
        default=timezone.now,
        db_index=True,
        help_text=_("When was this FAT link created"),
    )

    fleet = models.CharField(
        max_length=254,
        blank=False,
        default=None,
        help_text=_("The FAT link fleet name"),
    )

    hash = models.CharField(
        max_length=254, db_index=True, unique=True, help_text=_("The FAT link hash")
    )

    creator = models.ForeignKey(
        to=User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
        help_text=_("Who created the FAT link?"),
    )

    character = models.ForeignKey(
        to=EveCharacter,
        related_name="+",
        on_delete=models.CASCADE,
        default=None,
        null=True,
        help_text=_("Character this FAT link has been created with"),
    )

    link_type = models.ForeignKey(
        to=FleetType,
        related_name="+",
        on_delete=models.CASCADE,
        null=True,
        help_text="Deprecated setting, will be removed in a future version …",
    )

    fleet_type = models.CharField(
        blank=True,
        default="",
        max_length=254,
        help_text=_("The FAT link fleet type, if it's set"),
    )

    doctrine = models.CharField(
        blank=True, default="", max_length=254, help_text=_("The FAT link doctrine")
    )

    is_esilink = models.BooleanField(
        default=False, help_text=_("Whether this FAT link was created via ESI or not")
    )

    is_registered_on_esi = models.BooleanField(
        default=False,
        help_text=_("Whether the fleet to this FAT link is available in ESI or not"),
    )

    esi_fleet_id = models.BigIntegerField(blank=True, null=True)

    reopened = models.BooleanField(
        default=False, help_text=_("Has this FAT link being re-opened?")
    )

    last_esi_error = models.CharField(
        max_length=15, blank=True, default="", choices=EsiError.choices
    )

    last_esi_error_time = models.DateTimeField(null=True, blank=True, default=None)

    esi_error_count = models.IntegerField(default=0)

    objects: ClassVar[FatLinkManager] = FatLinkManager()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        ordering = ("-created",)
        verbose_name = _("FAT link")
        verbose_name_plural = _("FAT links")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return f"{self.fleet} - {self.hash}"

    @transaction.atomic()
    def save(self, *args, **kwargs):
        """
        Add the hash on save

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        """

        try:
            self.hash
        except ObjectDoesNotExist:
            self.hash = get_hash_on_save()
        super().save(*args, **kwargs)

    @property
    def number_of_fats(self):
        """
        Returns the number of registered FATs

        :return:
        :rtype:
        """

        return self.afat_fats.count()


class Duration(models.Model):
    """
    FAT link duration (expiry time in minutes)
    """

    duration = models.PositiveIntegerField()
    fleet = models.ForeignKey(
        to=FatLink, related_name="duration", on_delete=models.CASCADE
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("FAT link duration")
        verbose_name_plural = _("FAT link durations")


# AFat Model
class Fat(models.Model):
    """
    AFat
    """

    character = models.ForeignKey(
        to=EveCharacter,
        related_name="afat_fats",
        on_delete=models.CASCADE,
        help_text=_("Character who registered this FAT"),
    )

    corporation_eve_id = models.BigIntegerField(
        null=True,
        help_text=_(
            "Corporation EVE ID of the character who registered this FAT at the time of registration"
        ),
    )

    alliance_eve_id = models.BigIntegerField(
        null=True,
        help_text=_(
            "Alliance EVE ID of the character who registered this FAT at the time of registration"
        ),
    )

    fatlink = models.ForeignKey(
        to=FatLink,
        related_name="afat_fats",
        on_delete=models.CASCADE,
        help_text=_("The FAT link the character registered at"),
    )

    system = models.CharField(
        max_length=100, null=True, help_text=_("The system the character is in")
    )

    shiptype = models.CharField(
        max_length=100,
        null=True,
        db_index=True,
        help_text=_("The ship the character was flying"),
    )

    objects: ClassVar[FatManager] = FatManager()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        AFat :: Meta
        """

        default_permissions = ()
        unique_together = (("character", "fatlink"),)
        verbose_name = _("FAT")
        verbose_name_plural = _("FATs")

    def __str__(self) -> str:
        """
        Return the objects string name

        :return:
        :rtype:
        """

        return f"{self.fatlink} - {self.character}"


# AFat Log Model
class Log(models.Model):
    """
    The log
    """

    class Event(models.TextChoices):
        """
        Choices for log event
        """

        CREATE_FATLINK = "CR_FAT_LINK", _("FAT link created")
        CHANGE_FATLINK = "CH_FAT_LINK", _("FAT link changed")
        DELETE_FATLINK = "RM_FAT_LINK", _("FAT link removed")
        REOPEN_FATLINK = "RO_FAT_LINK", _("FAT link re-opened")
        # CREATE_FAT = "CR_FAT", _("FAT registered")
        DELETE_FAT = "RM_FAT", _("FAT removed")
        MANUAL_FAT = "CR_FAT_MAN", _("Manual FAT added")

    log_time = models.DateTimeField(default=timezone.now, db_index=True)
    user = models.ForeignKey(
        to=User,
        related_name="afat_log",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(value=get_sentinel_user),
    )
    log_event = models.CharField(
        max_length=11,
        blank=False,
        choices=Event.choices,
        default=Event.CREATE_FATLINK,
    )
    log_text = models.TextField()
    fatlink_hash = models.CharField(max_length=254)

    class Meta:  # pylint: disable=too-few-public-methods
        """
        AFatLog :: Meta
        """

        default_permissions = ()
        verbose_name = _("Log")
        verbose_name_plural = _("Logs")


class Doctrine(models.Model):
    """
    Fleet Doctrines
    """

    # Doctrine name
    name = models.CharField(
        max_length=255,
        unique=True,
        help_text=_("Short name to identify this doctrine"),
        verbose_name=_("Name"),
    )

    # Doctrine notes
    notes = models.TextField(
        default="",
        blank=True,
        help_text=_(
            "You can add notes about this doctrine here if you want. (optional)"
        ),
        verbose_name=_("Notes"),
    )

    # Is doctrine active
    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=_("Whether this doctrine is enabled or not."),
        verbose_name=_("Is enabled"),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        FleetDoctrine :: Meta
        """

        verbose_name = _("Doctrine")
        verbose_name_plural = _("Doctrines")
        default_permissions = ()

    def __str__(self) -> str:
        """
        String representation of the object

        :return:
        :rtype:
        """

        return str(self.name)


class Setting(SingletonModel):
    """
    Default forum settings
    """

    class Field(models.TextChoices):
        """
        Choices for Setting.Field
        """

        DEFAULT_FATLINK_EXPIRY_TIME = "default_fatlink_expiry_time", _(
            "Default FAT link expiry time"
        )
        DEFAULT_FATLINK_REOPEN_DURATION = "default_fatlink_reopen_duration", _(
            "Default FAT link reopen duration"
        )
        DEFAULT_FATLINK_REOPEN_GRACE_TIME = "default_fatlink_reopen_grace_time", _(
            "Default FAT link reopen grace time"
        )
        DEFAULT_LOG_DURATION = "default_log_duration", _("Default log duration")
        USE_DOCTRINES_FROM_FITTINGS_MODULE = "use_doctrines_from_fittings_module", _(
            "Use doctrines from fittings module"
        )

    default_fatlink_expiry_time = models.PositiveIntegerField(
        default=60,
        help_text=_(
            "Default expiry time for clickable FAT links in minutes. "
            "(Default: 60 minutes)"
        ),
        verbose_name=Field.DEFAULT_FATLINK_EXPIRY_TIME.label,  # pylint: disable=no-member
    )

    default_fatlink_reopen_grace_time = models.PositiveIntegerField(
        default=60,
        help_text=_(
            "Default time in minutes a FAT link can be re-opened after it is expired. "
            "(Default: 60 minutes)"
        ),
        verbose_name=Field.DEFAULT_FATLINK_REOPEN_GRACE_TIME.label,  # pylint: disable=no-member
    )

    default_fatlink_reopen_duration = models.PositiveIntegerField(
        default=60,
        help_text=_(
            "Default time in minutes a FAT link is re-opened for. "
            "(Default: 60 minutes)"
        ),
        verbose_name=Field.DEFAULT_FATLINK_REOPEN_DURATION.label,  # pylint: disable=no-member
    )

    default_log_duration = models.PositiveIntegerField(
        default=60,
        help_text=_("Default time in days a log entry is kept. (Default: 60 days)"),
        verbose_name=Field.DEFAULT_LOG_DURATION.label,  # pylint: disable=no-member
    )

    use_doctrines_from_fittings_module = models.BooleanField(
        default=False,
        db_index=True,
        help_text=_(
            "Whether to use the doctrines from the Fittings modules in the doctrine "
            "dropdown. Note: The fittings module needs to be installed for this."
        ),
        verbose_name=Field.USE_DOCTRINES_FROM_FITTINGS_MODULE.label,  # pylint: disable=no-member
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Setting :: Meta
        """

        default_permissions = ()
        verbose_name = _("Setting")
        verbose_name_plural = _("Settings")

    def __str__(self) -> str:
        """
        String representation of the object

        :return:
        :rtype:
        """

        return str(_("AFAT Settings"))

    @staticmethod
    def get_setting(setting_key: str) -> Any:
        """
        Get the setting value for a given setting key

        :param setting_key: The setting key
        :type setting_key: str
        :return: The setting value
        :rtype: Any
        """

        try:
            return getattr(Setting.get_solo(), setting_key)
        except AttributeError as exc:
            raise KeyError(f"Setting key '{setting_key}' does not exist.") from exc



"""
Dashboard related views
"""

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.evecharacter import get_user_from_evecharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__
from afat.helper.association import get_all_characters_with_fats_from_user
from afat.helper.views import convert_fatlinks_to_dict, convert_fats_to_dict
from afat.models import Fat, FatLink

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required()
@permission_required(perm="afat.basic_access")
def overview(request: WSGIRequest) -> HttpResponse:
    """
    Dashboard view

    :param request:
    :type request:
    :return:
    :rtype:
    """

    characters = get_all_characters_with_fats_from_user(request.user)

    context = {"characters": characters}

    logger.info(msg=f"Module called by {request.user}")

    return render(
        request=request,
        template_name="afat/view/dashboard/dashboard.html",
        context=context,
    )


@login_required
@permission_required(perm="afat.basic_access")
def ajax_recent_get_fats_by_character(
    request: WSGIRequest, charid: int
) -> JsonResponse:
    """
    Ajax call :: get all FATs for a given character

    :param request:
    :type request:
    :param charid:
    :type charid:
    :return:
    :rtype:
    """

    character = EveCharacter.objects.get(character_id=charid)
    user_by_character = get_user_from_evecharacter(character=character)

    # Check if the user has access to the character
    if user_by_character == request.user:
        fats = (
            Fat.objects.select_related_default()
            .filter(character=character)
            .order_by("fatlink__created")
            .reverse()[:10]
        )

        character_fat_rows = [
            convert_fats_to_dict(request=request, fat=fat) for fat in fats
        ]

        return JsonResponse(data=character_fat_rows, safe=False)

    # If the user does not have access to the character
    return JsonResponse(data=[], safe=False)


@login_required
@permission_required(perm="afat.basic_access")
def ajax_get_recent_fatlinks(request: WSGIRequest) -> JsonResponse:
    """
    Ajax call :: get recent fat links for the dashboard datatable

    :param request:
    :type request:
    :return:
    :rtype:
    """

    fatlink_ids = list(
        FatLink.objects.select_related_default()
        .order_by("-created")[:10]
        .values_list("id", flat=True)
    )

    fatlinks = (
        FatLink.objects.filter(id__in=fatlink_ids)
        .order_by("-created")
        .annotate_fats_count()  # pylint: disable=duplicate-code
    )

    fatlink_rows = [
        convert_fatlinks_to_dict(
            request=request,
            fatlink=fatlink,
            close_esi_redirect=reverse(viewname="afat:dashboard"),
        )
        for fatlink in fatlinks
    ]

    return JsonResponse(data=fatlink_rows, safe=False)



"""
Test checks for installed modules we might use
"""

# Django
from django.test import modify_settings

# Alliance Auth AFAT
from afat.app_settings import (
    fittings_installed,
    securegroups_installed,
    use_fittings_module_for_doctrines,
)
from afat.models import Setting
from afat.tests import BaseTestCase


class TestModulesInstalled(BaseTestCase):
    """
    Test if modules are installed
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()

    @modify_settings(INSTALLED_APPS={"remove": "fittings"})
    def test_for_fittings_installed_when_not_installed(self):
        """
        Test for fittings_installed when it is not installed.

        :return:
        :rtype:
        """

        self.assertFalse(expr=fittings_installed())

    @modify_settings(INSTALLED_APPS={"append": "fittings"})
    def test_for_fittings_installed_when_installed(self):
        """
        Test for fittings_installed when it is installed.

        :return:
        :rtype:
        """

        self.assertTrue(expr=fittings_installed())

    @modify_settings(INSTALLED_APPS={"remove": "securegroups"})
    def test_for_securegroups_installed_when_not_installed(self):
        """
        Test for securegroups when it is not installed.

        :return:
        :rtype:
        """

        self.assertFalse(expr=securegroups_installed())

    @modify_settings(INSTALLED_APPS={"append": "securegroups"})
    def test_for_securegroups_installed_when_installed(self):
        """
        Test for securegroups when it is installed.

        :return:
        :rtype:
        """

        self.assertTrue(expr=securegroups_installed())

    @modify_settings(INSTALLED_APPS={"remove": "fittings"})
    def test_for_use_fittings_module_for_doctrines_when_fittings_not_installed_and_not_enabled(
        self,
    ):
        """
        Test for use_fittings_module_for_doctrines when the fittings module is not installed and not enabled in settings.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()
        settings.use_doctrines_from_fittings_module = False
        settings.save()

        self.assertFalse(
            expr=Setting.get_setting(Setting.Field.USE_DOCTRINES_FROM_FITTINGS_MODULE)
        )
        self.assertFalse(expr=use_fittings_module_for_doctrines())

    @modify_settings(INSTALLED_APPS={"remove": "fittings"})
    def test_for_use_fittings_module_for_doctrines_when_fittings_not_installed_but_enabled(
        self,
    ):
        """
        Test for use_fittings_module_for_doctrines when the fittings module is not installed but enabled in settings.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()
        settings.use_doctrines_from_fittings_module = True
        settings.save()

        self.assertTrue(
            expr=Setting.get_setting(Setting.Field.USE_DOCTRINES_FROM_FITTINGS_MODULE)
        )
        self.assertFalse(expr=use_fittings_module_for_doctrines())

    @modify_settings(INSTALLED_APPS={"append": "fittings"})
    def test_for_use_fittings_module_for_doctrines_when_fittings_installed_and_enabled(
        self,
    ):
        """
        Test for use_fittings_module_for_doctrines when the fittings module is installed and enabled in settings.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()
        settings.use_doctrines_from_fittings_module = True
        settings.save()

        self.assertTrue(
            expr=Setting.get_setting(Setting.Field.USE_DOCTRINES_FROM_FITTINGS_MODULE)
        )
        self.assertTrue(expr=use_fittings_module_for_doctrines())

    @modify_settings(INSTALLED_APPS={"append": "fittings"})
    def test_for_use_fittings_module_for_doctrines_when_fittings_installed_but_not_enabled(
        self,
    ):
        """
        Test for use_fittings_module_for_doctrines when the fittings module is installed but not enabled in settings.

        :return:
        :rtype:
        """

        settings = Setting.get_solo()
        settings.use_doctrines_from_fittings_module = False
        settings.save()

        self.assertFalse(
            expr=Setting.get_setting(Setting.Field.USE_DOCTRINES_FROM_FITTINGS_MODULE)
        )
        self.assertFalse(expr=use_fittings_module_for_doctrines())



# Generated by Django 2.2.14 on 2020-08-20 20:13

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("afat", "0005_auto_20200816_2042"),
    ]

    operations = [
        migrations.CreateModel(
            name="AFatLinkType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=254)),
                ("deleted_at", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="afatlink",
            name="link_type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="afat.AFatLinkType",
            ),
        ),
    ]



"""
Versioned static URLs to break browser caches when changing the app version.
"""

# Django
from django.template.defaulttags import register
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# Alliance Auth AFAT
from afat import __title__

logger = LoggerAddTag(my_logger=get_extension_logger(__name__), prefix=__title__)


@register.filter
def month_name(month_number):
    """
    Template tag :: get month name from month number
    example: {{ event.month|month_name }}

    :param month_number:
    :return:
    """

    month_mapper = {
        1: _("January"),
        2: _("February"),
        3: _("March"),
        4: _("April"),
        5: _("May"),
        6: _("June"),
        7: _("July"),
        8: _("August"),
        9: _("September"),
        10: _("October"),
        11: _("November"),
        12: _("December"),
    }

    return month_mapper[int(month_number)]


@register.filter
def sum_values(dictionary):
    """
    Template tag :: sum all values in a dictionary
    example: {{ dictionary|sum_values }}

    :param dictionary:
    :type dictionary:
    :return:
    :rtype:
    """

    return sum(dictionary.values())



# Generated by Django 3.1.4 on 2020-12-26 20:01

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0012_index_additions"),
        ("afat", "0014_auto_20201224_0930"),
    ]

    operations = [
        migrations.AddField(
            model_name="afatlink",
            name="character",
            field=models.ForeignKey(
                default=None,
                help_text="Character this fatlink has been created with",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="eveonline.evecharacter",
            ),
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-30 12:26

# Django
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("afat", "0016_permissions_overhaul"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="afat",
            name="deleted_at",
        ),
        migrations.RemoveField(
            model_name="afatlink",
            name="deleted_at",
        ),
        migrations.RemoveField(
            model_name="afatlinktype",
            name="deleted_at",
        ),
        migrations.DeleteModel(
            name="AFatDelLog",
        ),
    ]



"""
Constants used in this module
"""

# Django
from django.utils.text import slugify

# Alliance Auth AFAT
from afat import __title__

APP_BASE_URL = slugify(value=__title__, allow_unicode=True)

INTERNAL_URL_PREFIX = "-"



# Generated by Django 3.2.4 on 2021-06-17 10:37

# Django
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

# Alliance Auth AFAT
import afat.models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0014_auto_20210105_1413"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("afat", "0019_add_several_indices"),
    ]

    operations = [
        migrations.AlterField(
            model_name="afat",
            name="afatlink",
            field=models.ForeignKey(
                help_text="The FAT link the character registered at",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afats",
                to="afat.afatlink",
            ),
        ),
        migrations.AlterField(
            model_name="afat",
            name="character",
            field=models.ForeignKey(
                help_text="Character who registered this FAT",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="afats",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="afattime",
            field=models.DateTimeField(
                db_index=True,
                default=django.utils.timezone.now,
                help_text="When was this FAT link created",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="character",
            field=models.ForeignKey(
                default=None,
                help_text="Character this FAT link has been created with",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="creator",
            field=models.ForeignKey(
                help_text="Who created the FAT link?",
                on_delete=models.SET(afat.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="fleet",
            field=models.CharField(
                help_text="The FAT link fleet name", max_length=254, null=True
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="hash",
            field=models.CharField(
                db_index=True,
                help_text="The FAT link hash",
                max_length=254,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="is_esilink",
            field=models.BooleanField(
                default=False,
                help_text="Whether this FAT link was created via ESI or not",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="is_registered_on_esi",
            field=models.BooleanField(
                default=False,
                help_text="Whether this is an ESI fat link is registered on ESI or not",
            ),
        ),
        migrations.AlterField(
            model_name="afatlink",
            name="link_type",
            field=models.ForeignKey(
                help_text="The FAT link fleet type, if it's set",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="afat.afatlinktype",
            ),
        ),
        migrations.AlterField(
            model_name="afatlinktype",
            name="is_enabled",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="Whether this fleet type is active or not",
            ),
        ),
    ]



"""
Migration script to migrate fleet types from the old format to the new format.
"""

# Django
from django.core.management import BaseCommand

# Alliance Auth AFAT
from afat.models import FatLink


def get_input(text) -> str:
    """
    Wrapped input to enable import

    :param text:
    :type text:
    :return:
    :rtype:
    """

    return input(text)


class Command(BaseCommand):
    """
    Initial import of FAT data from AA FAT module
    """

    help = (
        "Migrating fleet types."
        "This command will migrate the fleet types from the old format to the new format."
    )

    def _migrate_fleet_types(self) -> None:
        """
        Start the import

        :return:
        :rtype:
        """

        # Migrate fleet types to the new format
        fatlinks = FatLink.objects.all()

        for fatlink in fatlinks:
            fatlink.fleet_type = (
                str(fatlink.link_type)
                if fatlink.link_type and fatlink.fleet_type == ""
                else fatlink.fleet_type
            )
            fatlink.save()

        self.stdout.write(msg=self.style.SUCCESS("Migration complete!"))

    def handle(self, *args, **options):  # pylint: disable=unused-argument
        """
        Ask before running …

        :param args:
        :type args:
        :param options:
        :type options:
        :return:
        :rtype:
        """

        self.stdout.write(
            msg=self.style.SUCCESS(
                "This command will migrate the fleet types from the old format to the new format."
            )
        )

        user_input = get_input(text="Are you sure you want to proceed? (yes/no) ")

        if user_input == "yes":
            self.stdout.write(msg="Starting import. Please stand by.")
            self._migrate_fleet_types()
        else:
            self.stdout.write(msg=self.style.WARNING("Aborted."))



# Generated by Django 4.2.20 on 2025-04-11 14:36

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("afat", "0024_setting_doctrine_and_fleet_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="fat",
            name="alliance_eve_id",
            field=models.BigIntegerField(
                help_text="Alliance EVE ID of the character who registered this FAT at the time of registration",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="fat",
            name="corporation_eve_id",
            field=models.BigIntegerField(
                help_text="Corporation EVE ID of the character who registered this FAT at the time of registration",
                null=True,
            ),
        ),
    ]



"""
Test cases for the task in the afat module.
"""

# Standard Library
from datetime import timedelta
from unittest.mock import ANY, MagicMock, patch

# Third Party
import kombu

# Django
from django.utils.datetime_safe import datetime

# Alliance Auth (External Libs)
from app_utils.esi import EsiStatus

# Alliance Auth AFAT
from afat.models import FatLink
from afat.tasks import (
    _close_esi_fleet,
    _esi_fatlinks_error_handling,
    _process_esi_fatlink,
    logrotate,
    process_fats,
    update_esi_fatlinks,
)
from afat.tests import BaseTestCase


class TestLogrotateTask(BaseTestCase):
    """
    Test cases for the logrotate task.
    """

    @patch("afat.tasks.Setting.get_setting")
    @patch("afat.tasks.Log.objects.filter")
    def test_logrotate_removes_old_logs(self, mock_filter, mock_get_setting):
        """
        Test that the logrotate task removes logs older than the specified duration.

        :param mock_filter:
        :type mock_filter:
        :param mock_get_setting:
        :type mock_get_setting:
        :return:
        :rtype:
        """

        mock_get_setting.return_value = 30
        mock_filter.return_value.delete.return_value = None

        logrotate()

        mock_filter.assert_called_once_with(log_time__lte=ANY)
        mock_filter.return_value.delete.assert_called_once()

    @patch("afat.tasks.Setting.get_setting")
    @patch("afat.tasks.Log.objects.filter")
    def test_logrotate_handles_no_old_logs(self, mock_filter, mock_get_setting):
        """
        Test that the logrotate task handles the case where there are no old logs.

        :param mock_filter:
        :type mock_filter:
        :param mock_get_setting:
        :type mock_get_setting:
        :return:
        :rtype:
        """

        mock_get_setting.return_value = 30
        mock_filter.return_value.delete.return_value = None

        logrotate()

        mock_filter.assert_called_once_with(log_time__lte=ANY)
        mock_filter.return_value.delete.assert_called_once()


class UpdateEsiFatlinksTests(BaseTestCase):
    """
    Test cases for the update_esi_fatlinks task.
    """

    @patch("afat.tasks.fetch_esi_status")
    @patch("afat.tasks.logger")
    def test_checking_esi_fat_links_when_esi_is_offline(
        self, mock_logger, mock_fetch_esi_status
    ):
        """
        Test that the update_esi_fatlinks task handles the case when ESI is offline.

        :param mock_logger:
        :type mock_logger:
        :param mock_fetch_esi_status:
        :type mock_fetch_esi_status:
        :return:
        :rtype:
        """

        mock_fetch_esi_status.return_value = EsiStatus(is_online=False)
        update_esi_fatlinks()
        mock_logger.warning.assert_called_once_with(
            "ESI doesn't seem to be available at this time. Aborting."
        )

    @patch("afat.tasks.fetch_esi_status")
    @patch("afat.tasks.FatLink.objects.select_related_default")
    @patch("afat.tasks.logger")
    def test_checking_esi_fat_links_when_no_fatlinks(
        self, mock_logger, mock_fatlink_queryset, mock_fetch_esi_status
    ):
        """
        Test that the update_esi_fatlinks task handles the case when there are no ESI FAT links.

        :param mock_logger:
        :type mock_logger:
        :param mock_fatlink_queryset:
        :type mock_fatlink_queryset:
        :param mock_fetch_esi_status:
        :type mock_fetch_esi_status:
        :return:
        :rtype:
        """

        mock_fetch_esi_status.return_value = EsiStatus(is_online=True)
        mock_fatlink_queryset.return_value.filter.return_value.distinct.return_value = (
            []
        )
        update_esi_fatlinks()
        mock_logger.debug.assert_any_call(msg="Found 0 ESI FAT links to process")

    @patch("afat.tasks.fetch_esi_status")
    @patch("afat.tasks.FatLink.objects.select_related_default")
    @patch("afat.tasks.logger")
    @patch("afat.tasks._process_esi_fatlink")
    def test_checking_esi_fat_links_when_fatlinks_exist(
        self,
        mock_process_esi_fatlink,
        mock_logger,
        mock_fatlink_queryset,
        mock_fetch_esi_status,
    ):
        """
        Test that the update_esi_fatlinks task handles the case when there are ESI FAT links.

        :param mock_process_esi_fatlink:
        :type mock_process_esi_fatlink:
        :param mock_logger:
        :type mock_logger:
        :param mock_fatlink_queryset:
        :type mock_fatlink_queryset:
        :param mock_fetch_esi_status:
        :type mock_fetch_esi_status:
        :return:
        :rtype:
        """

        mock_fetch_esi_status.return_value = EsiStatus(is_online=True)
        mock_fatlink_queryset.return_value.filter.return_value.distinct.return_value = [
            MagicMock()
        ]
        update_esi_fatlinks()
        mock_process_esi_fatlink.assert_called_once()
        mock_logger.debug.assert_any_call(msg="Found 1 ESI FAT links to process")


class ProcessEsiFatlinkTests(BaseTestCase):
    """
    Test cases for the _process_esi_fatlink function.
    """

    @patch("afat.tasks.logger")
    @patch("afat.tasks._close_esi_fleet")
    def test_processing_esi_fatlink_when_no_creator(
        self, mock_close_esi_fleet, mock_logger
    ):
        """
        Test that the _process_esi_fatlink function handles the case when there is no creator.

        :param mock_close_esi_fleet:
        :type mock_close_esi_fleet:
        :param mock_logger:
        :type mock_logger:
        :return:
        :rtype:
        """

        fatlink = MagicMock()
        fatlink.creator.profile.main_character = None
        _process_esi_fatlink(fatlink)
        mock_close_esi_fleet.assert_called_once_with(
            fatlink=fatlink, reason="No FAT link creator available."
        )

    @patch("afat.tasks.logger")
    @patch("afat.tasks._check_for_esi_fleet")
    def test_processing_esi_fatlink_when_no_fleet(
        self, mock_check_for_esi_fleet, mock_logger
    ):
        """
        Test that the _process_esi_fatlink function handles the case when there is no fleet.

        :param mock_check_for_esi_fleet:
        :type mock_check_for_esi_fleet:
        :param mock_logger:
        :type mock_logger:
        :return:
        :rtype:
        """

        fatlink = MagicMock()
        fatlink.creator.profile.main_character = MagicMock()
        mock_check_for_esi_fleet.return_value = None
        _process_esi_fatlink(fatlink)
        mock_check_for_esi_fleet.assert_called_once_with(fatlink=fatlink)


class EsiFatlinksErrorHandlingTests(BaseTestCase):
    """
    Test cases for the _esi_fatlinks_error_handling function.
    """

    @patch("afat.tasks.timezone.now")
    @patch("afat.tasks._close_esi_fleet")
    def test_handles_error_within_grace_period(self, mock_close_fleet, mock_now):
        """
        Test that the _esi_fatlinks_error_handling function handles the case when an error occurs within the grace period.

        :param mock_close_fleet:
        :type mock_close_fleet:
        :param mock_now:
        :type mock_now:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        error_key = MagicMock()
        error_key.label = "Test Error"
        now = datetime(2023, 10, 1, 12, 0, 0)
        mock_now.return_value = now
        fatlink.last_esi_error = error_key
        fatlink.last_esi_error_time = now - timedelta(seconds=30)
        fatlink.esi_error_count = 3

        _esi_fatlinks_error_handling(error_key, fatlink)

        mock_close_fleet.assert_called_once_with(
            fatlink=fatlink, reason=error_key.label
        )
        fatlink.save.assert_not_called()

    @patch("afat.tasks.timezone.now")
    def test_increments_error_count(self, mock_now):
        """
        Test that the _esi_fatlinks_error_handling function increments the error count.

        :param mock_now:
        :type mock_now:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        error_key = MagicMock()
        error_key.label = "Test Error"
        now = datetime(2023, 10, 1, 12, 0, 0)
        mock_now.return_value = now
        fatlink.last_esi_error = error_key
        fatlink.last_esi_error_time = now - timedelta(seconds=30)
        fatlink.esi_error_count = 2

        _esi_fatlinks_error_handling(error_key, fatlink)

        self.assertEqual(fatlink.esi_error_count, 3)
        fatlink.save.assert_called_once()

    @patch("afat.tasks.timezone.now")
    def test_resets_error_count_after_grace_period(self, mock_now):
        """
        Test that the _esi_fatlinks_error_handling function resets the error count after the grace period.

        :param mock_now:
        :type mock_now:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        error_key = MagicMock()
        error_key.label = "Test Error"
        now = datetime(2023, 10, 1, 12, 0, 0)
        mock_now.return_value = now
        fatlink.last_esi_error = error_key
        fatlink.last_esi_error_time = now - timedelta(seconds=100)
        fatlink.esi_error_count = 2

        _esi_fatlinks_error_handling(error_key, fatlink)

        self.assertEqual(fatlink.esi_error_count, 1)
        fatlink.save.assert_called_once()

    @patch("afat.tasks.timezone.now")
    def test_handles_new_error(self, mock_now):
        """
        Test that the _esi_fatlinks_error_handling function handles a new error.

        :param mock_now:
        :type mock_now:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        error_key = MagicMock()
        error_key.label = "Test Error"
        now = datetime(2023, 10, 1, 12, 0, 0)
        mock_now.return_value = now
        fatlink.last_esi_error = None
        fatlink.last_esi_error_time = None
        fatlink.esi_error_count = 0

        _esi_fatlinks_error_handling(error_key, fatlink)

        self.assertEqual(fatlink.esi_error_count, 1)
        self.assertEqual(fatlink.last_esi_error, error_key)
        self.assertEqual(fatlink.last_esi_error_time, mock_now.return_value)
        fatlink.save.assert_called_once()


class CloseEsiFleetTests(BaseTestCase):
    """
    Test cases for the _close_esi_fleet function.
    """

    @patch("afat.tasks.logger.info")
    def test_closes_fleet_successfully(self, mock_logger_info):
        """
        Test that the _close_esi_fleet function closes the fleet successfully.

        :param mock_logger_info:
        :type mock_logger_info:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        fatlink.hash = "test_hash"

        _close_esi_fleet(fatlink=fatlink, reason="Test Reason")

        fatlink.is_registered_on_esi = False
        fatlink.save.assert_called_once()
        mock_logger_info.assert_called_once_with(
            msg='Closing ESI FAT link with hash "test_hash". Reason: Test Reason'
        )

    @patch("afat.tasks.logger.info")
    def test_handles_empty_reason(self, mock_logger_info):
        """
        Test that the _close_esi_fleet function handles an empty reason.

        :param mock_logger_info:
        :type mock_logger_info:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        fatlink.hash = "test_hash"

        _close_esi_fleet(fatlink=fatlink, reason="")

        fatlink.is_registered_on_esi = False
        fatlink.save.assert_called_once()
        mock_logger_info.assert_called_once_with(
            msg='Closing ESI FAT link with hash "test_hash". Reason: '
        )

    @patch("afat.tasks.logger.info")
    def test_handles_none_reason(self, mock_logger_info):
        """
        Test that the _close_esi_fleet function handles a None reason.

        :param mock_logger_info:
        :type mock_logger_info:
        :return:
        :rtype:
        """

        fatlink = MagicMock(spec=FatLink)
        fatlink.hash = "test_hash"

        _close_esi_fleet(fatlink=fatlink, reason=None)

        fatlink.is_registered_on_esi = False
        fatlink.save.assert_called_once()
        mock_logger_info.assert_called_once_with(
            msg='Closing ESI FAT link with hash "test_hash". Reason: None'
        )


class ProcessFatsTests(BaseTestCase):
    """
    Test cases for the process_fats function.
    """

    @patch("afat.tasks.process_character.si")
    @patch("afat.tasks.group")
    def test_processes_fat_link_data_from_esi(
        self, mock_group, mock_process_character_si
    ):
        """
        Test that the process_fats function processes FAT link data from ESI.

        :param mock_group:
        :type mock_group:
        :param mock_process_character_si:
        :type mock_process_character_si:
        :return:
        :rtype:
        """

        data_list = [
            {"character_id": 1, "solar_system_id": 100, "ship_type_id": 200},
            {"character_id": 2, "solar_system_id": 101, "ship_type_id": 201},
        ]
        fatlink_hash = "test_hash"
        mock_group.return_value.delay = MagicMock()

        process_fats(data_list, "esi", fatlink_hash)

        self.assertEqual(mock_process_character_si.call_count, 2)
        mock_group.assert_called_once()
        mock_group.return_value.delay.assert_called_once()

    @patch("afat.tasks.process_character.si")
    @patch("afat.tasks.group")
    def test_processes_fat_link_data_with_no_tasks(
        self, mock_group, mock_process_character_si
    ):
        """
        Test that the process_fats function handles the case when there are no tasks to process.

        :param mock_group:
        :type mock_group:
        :param mock_process_character_si:
        :type mock_process_character_si:
        :return:
        :rtype:
        """

        data_list = []
        fatlink_hash = "test_hash"

        process_fats(data_list, "esi", fatlink_hash)

        mock_process_character_si.assert_not_called()
        mock_group.assert_not_called()

    @patch("afat.tasks.process_character.si")
    @patch("afat.tasks.group")
    def test_handles_kombu_encode_error(self, mock_group, mock_process_character_si):
        data_list = [
            {"character_id": 1, "solar_system_id": 100, "ship_type_id": 200},
        ]
        fatlink_hash = "test_hash"
        mock_group.return_value.delay.side_effect = kombu.exceptions.EncodeError

        process_fats(data_list, "esi", fatlink_hash)

        self.assertEqual(mock_process_character_si.call_count, 1)
        mock_group.assert_called_once()
        mock_group.return_value.delay.assert_called_once()



"""
Test logs view
"""

# Standard Library
from http import HTTPStatus

# Third Party
from pytz import utc

# Django
from django.urls import reverse
from django.utils.datetime_safe import datetime

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# Alliance Auth AFAT
from afat.models import FatLink
from afat.tests import BaseTestCase
from afat.tests.fixtures.load_allianceauth import load_allianceauth

MODULE_PATH = "afat.views.logs"


class TestLogsView(BaseTestCase):
    """
    Test the logs view
    """

    @classmethod
    def setUpClass(cls):
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()
        load_allianceauth()

        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_without_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1002.character_id,
            permissions=["afat.basic_access"],
        )

        cls.user_with_manage_afat, _ = create_user_from_evecharacter(
            character_id=cls.character_1003.character_id,
            permissions=["afat.basic_access", "afat.manage_afat"],
        )

        cls.user_with_log_view, _ = create_user_from_evecharacter(
            character_id=cls.character_1004.character_id,
            permissions=["afat.basic_access", "afat.log_view"],
        )

        # Generate some FAT links and FATs
        cls.afat_link_april_1 = FatLink.objects.create(
            fleet="April Fleet 1",
            hash="1231",
            creator=cls.user_with_manage_afat,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=1, tzinfo=utc),
        )
        cls.afat_link_april_2 = FatLink.objects.create(
            fleet="April Fleet 2",
            hash="1232",
            creator=cls.user_with_manage_afat,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=15, tzinfo=utc),
        )

    def test_should_not_show_log_view_for_user_without_access(self):
        """
        Test should not show log view for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_without_access)

        url = reverse(viewname="afat:logs_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_not_show_log_view_for_user_with_basic_access(self):
        """
        Test should not show log view for user with basic access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:logs_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_show_log_view_for_user_with_manage_afat_permission(self):
        """
        Test should show log view for user with manage afat permission

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(viewname="afat:logs_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_log_view_for_user_with_log_view_permission(self):
        """
        Test should show log view for user with log view permission

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_log_view)

        url = reverse(viewname="afat:logs_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_ajax_get_logs(self):
        """
        Test ajax get logs

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_log_view)

        url = reverse(viewname="afat:logs_ajax_get_logs")
        result = self.client.get(path=url)

        self.assertEqual(first=result.status_code, second=HTTPStatus.OK)



# Standard Library
from http import HTTPStatus

# Third Party
from pytz import utc

# Django
from django.urls import reverse
from django.utils.datetime_safe import datetime

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

# Alliance Auth AFAT
from afat.models import Fat, FatLink
from afat.tests import BaseTestCase
from afat.tests.fixtures.load_allianceauth import load_allianceauth
from afat.tests.fixtures.utils import RequestStub
from afat.views.statistics import _calculate_year_stats

MODULE_PATH = "afat.views.statistics"


def response_content_to_str(response) -> str:
    """
    Convert response content to string

    :param response:
    :type response:
    :return:
    :rtype:
    """

    return response.content.decode(response.charset)


class TestStatistics(BaseTestCase):
    """
    Test the statistics views
    """

    @classmethod
    def setUpClass(cls):
        """
        Setup the test class

        :return:
        :rtype:
        """

        super().setUpClass()
        load_allianceauth()

        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_without_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1001.character_id
        )

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            character_id=cls.character_1002.character_id,
            permissions=["afat.basic_access"],
        )

        add_character_to_user(
            user=cls.user_with_basic_access, character=cls.character_1101
        )

        cls.user_with_manage_afat, _ = create_user_from_evecharacter(
            character_id=cls.character_1003.character_id,
            permissions=["afat.basic_access", "afat.manage_afat"],
        )

        cls.user_with_stats_corporation_other, _ = create_user_from_evecharacter(
            character_id=cls.character_1004.character_id,
            permissions=["afat.basic_access", "afat.stats_corporation_other"],
        )

        cls.user_with_stats_corporation_own, _ = create_user_from_evecharacter(
            character_id=cls.character_1005.character_id,
            permissions=["afat.basic_access", "afat.stats_corporation_own"],
        )

        # Generate some FAT links and FATs
        afat_link_april_1 = FatLink.objects.create(
            fleet="April Fleet 1",
            hash="1231",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=1, tzinfo=utc),
        )
        afat_link_april_2 = FatLink.objects.create(
            fleet="April Fleet 2",
            hash="1232",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=4, day=15, tzinfo=utc),
        )
        afat_link_september = FatLink.objects.create(
            fleet="September Fleet",
            hash="1233",
            creator=cls.user_with_basic_access,
            character=cls.character_1001,
            created=datetime(year=2020, month=9, day=1, tzinfo=utc),
        )

        Fat.objects.create(
            character=cls.character_1101, fatlink=afat_link_april_1, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1001, fatlink=afat_link_april_1, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1002, fatlink=afat_link_april_1, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1003, fatlink=afat_link_april_1, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1004, fatlink=afat_link_april_1, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1005, fatlink=afat_link_april_1, shiptype="Omen"
        )

        Fat.objects.create(
            character=cls.character_1101, fatlink=afat_link_april_2, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1004, fatlink=afat_link_april_2, shiptype="Thorax"
        )
        Fat.objects.create(
            character=cls.character_1002, fatlink=afat_link_april_2, shiptype="Thorax"
        )
        Fat.objects.create(
            character=cls.character_1003, fatlink=afat_link_april_2, shiptype="Omen"
        )

        Fat.objects.create(
            character=cls.character_1001, fatlink=afat_link_september, shiptype="Omen"
        )
        Fat.objects.create(
            character=cls.character_1004,
            fatlink=afat_link_september,
            shiptype="Guardian",
        )
        Fat.objects.create(
            character=cls.character_1005, fatlink=afat_link_september, shiptype="Omen"
        )

    def test_should_only_show_my_chars_and_only_those_with_fat_links(self):
        """
        Test that the overview page only shows the characters of the user that have FAT links.

        :return:
        :rtype:
        """

        result = _calculate_year_stats(
            request=RequestStub(user=self.user_with_basic_access), year=2020
        )

        self.assertDictEqual(
            d1=result,
            d2={
                "total": {4: 4},
                "characters": [
                    ("Clark Kent", {4: 2}, 1002),
                    ("Lex Luther", {4: 2}, 1101),
                ],
            },
        )

    def test_should_show_statistics_dashboard(self):
        """
        Test should show statistics dashboard

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:statistics_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_statistics_dashboard_for_year(self):
        """
        Test should show statistics dashboard for year

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(viewname="afat:statistics_overview", kwargs={"year": 2020})
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_statistics_dashboard_for_user_with_stats_corporation_other(
        self,
    ):
        """
        Test should show statistics dashboard for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(viewname="afat:statistics_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_statistics_dashboard_for_user_with_stats_corporation_own(self):
        """
        Test should show statistics dashboard for user with stats_corporation_own

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_own)

        url = reverse(viewname="afat:statistics_overview")
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_statistics_dashboard_for_user_without_access(self):
        """
        Test should not show statistics dashboard for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_without_access)

        url = reverse(viewname="afat:statistics_overview")
        res = self.client.get(path=url)

        self.assertNotEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=res.url,
            second="/account/login/?next=/fleet-activity-tracking/statistics/",
        )

    def test_should_show_own_character_stats(self):
        """
        Test should show own character stats

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_character",
            kwargs={
                "charid": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_other_character_stats_for_user_with_stats_corporation_own(
        self,
    ):
        """
        Test should show other character stats for user with stats_corporation_own

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_own)

        url = reverse(
            viewname="afat:statistics_character",
            kwargs={
                "charid": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_other_character_stats_for_user_with_stats_corporation_other(
        self,
    ):
        """
        Test should show other character stats for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_character",
            kwargs={
                "charid": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_other_character_stats_for_user_with_manage_afat(self):
        """
        Test should show other character stats for user with manage_afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_character",
            kwargs={
                "charid": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_other_character_stats_for_user(self):
        """
        Test should not show other character stats for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_character",
            kwargs={
                "charid": self.user_with_stats_corporation_other.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_show_own_corp_stats_for_user_with_stats_corporation_own(self):
        """
        Test should show own corp stats for user with stats_corporation_own

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_own)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={
                "corpid": self.user_with_stats_corporation_own.profile.main_character.corporation_id
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_other_corp_stats_for_user_with_stats_corporation_other(self):
        """
        Test should show other corp stats for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={
                "corpid": self.user_with_basic_access.profile.main_character.corporation_id
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_other_corp_stats_for_user_with_manage_afat(self):
        """
        Test should show other corp stats for user with manage_afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={
                "corpid": self.user_with_basic_access.profile.main_character.corporation_id
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_own_corp_stats_for_user(self):
        """
        Test should not show own corp stats for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={
                "corpid": self.user_with_basic_access.profile.main_character.corporation_id
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_not_show_other_corp_stats_for_user(self):
        """
        Test should not show other corp stats for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={
                "corpid": self.user_with_stats_corporation_other.profile.main_character.corporation_id
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_show_all_corp_stats_for_user_with_stats_corporation_other(self):
        """
        Test should show all corp stats for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={"corpid": 2002, "year": 2020},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_corp_stats_with_month_for_user_with_stats_corporation_other(
        self,
    ):
        """
        Test should show all corp stats with month for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_corporation",
            kwargs={"corpid": 2002, "year": 2020, "month": 4},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_all_corp_stats_for_user_with_stats_corporation_own(self):
        """
        Test should not show all corp stats for user with stats_corporation_own

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_own)

        url = reverse(viewname="afat:statistics_corporation", kwargs={"corpid": 2002})
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second="/fleet-activity-tracking/")

    def test_should_show_all_alliance_stats_with_for_user_with_stats_corporation_other(
        self,
    ):
        """
        Test should show all alliance stats for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_alliance_stats_with_for_user_with_manage_afat(
        self,
    ):
        """
        Test should show all alliance stats for user with manage_afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_alliance_stats_with_year_for_user_with_stats_corporation_other(
        self,
    ):
        """
        Test should show all alliance stats with year for user with stats_corporation_other

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001, "year": 2020},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_alliance_stats_with_year_for_user_with_manage_afat(
        self,
    ):
        """
        Test should show all alliance stats with year for user with manage_afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001, "year": 2020},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_alliance_stats_with_month_for_user_with_stats_corporation_other(
        self,
    ):
        # given
        self.client.force_login(user=self.user_with_stats_corporation_other)

        # when
        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001, "year": 2020, "month": 4},
        )
        res = self.client.get(path=url)

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_all_alliance_stats_with_month_for_user_with_manage_afat(
        self,
    ):
        """
        Test should show all alliance stats with month for user with manage_afat

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001, "year": 2020, "month": 4},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_all_alliance_stats_for_user(
        self,
    ):
        """
        Test should not show all alliance stats for user without access

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_alliance",
            kwargs={"allianceid": 3001},
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)

    def test_should_show_main_details_for_user_with_manage_perms(self):
        """
        Test that a user with the required permissions can access the view

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_manage_afat)

        url = reverse(
            viewname="afat:statistics_ajax_get_monthly_fats_for_main_character",
            kwargs={
                "corporation_id": self.user_with_basic_access.profile.main_character.corporation_id,
                "character_id": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_main_details_for_user_with_corporation_other_perms(self):
        """
        Test that a user with the required permissions can access the view (corporation_other)

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_stats_corporation_other)

        url = reverse(
            viewname="afat:statistics_ajax_get_monthly_fats_for_main_character",
            kwargs={
                "corporation_id": self.user_with_basic_access.profile.main_character.corporation_id,
                "character_id": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_not_show_main_details_for_user_without_perms(self):
        """
        Test that a user without the required permissions cannot access the view

        :return:
        :rtype:
        """

        self.client.force_login(user=self.user_with_basic_access)

        url = reverse(
            viewname="afat:statistics_ajax_get_monthly_fats_for_main_character",
            kwargs={
                "corporation_id": self.user_with_basic_access.profile.main_character.corporation_id,
                "character_id": self.user_with_basic_access.profile.main_character.character_id,
                "year": 2020,
                "month": 4,
            },
        )
        res = self.client.get(path=url)

        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=response_content_to_str(response=res),
            second="",
        )



"""
Helper functions for fat links view
"""

# Django
from django.db.models import Prefetch

# Alliance Auth
from allianceauth.authentication.admin import User

# Alliance Auth AFAT
from afat.app_settings import use_fittings_module_for_doctrines
from afat.models import FatLink


def get_esi_fleet_information_by_user(
    user: User,
) -> dict[str, bool | list[dict[int, FatLink]]]:
    """
    Get ESI fleet information by a given FC (user)

    :param user:
    :type user:
    :return:
    :rtype:
    """

    has_open_esi_fleets = False
    open_esi_fleets_list = []
    open_esi_fleets = (
        FatLink.objects.select_related_default()
        .filter(creator=user, is_esilink=True, is_registered_on_esi=True)
        .order_by("character__character_name")
    )

    if open_esi_fleets.count() > 0:
        has_open_esi_fleets = True

        for open_esi_fleet in open_esi_fleets:
            open_esi_fleets_list.append(open_esi_fleet)

    return {
        "has_open_esi_fleets": has_open_esi_fleets,
        "open_esi_fleets_list": open_esi_fleets_list,
    }


def get_doctrines() -> list[dict[int, str]]:
    """
    Get all enabled doctrines

    :return:
    :rtype:
    """

    if use_fittings_module_for_doctrines() is True:
        # Third Party
        from fittings.models import (  # pylint: disable=import-outside-toplevel
            Doctrine,
            Fitting,
        )

        cls = Doctrine.objects

        doctrines = (
            cls.prefetch_related("category")
            .prefetch_related(
                Prefetch(
                    lookup="fittings",
                    queryset=Fitting.objects.select_related("ship_type"),
                )
            )
            .union(
                cls.prefetch_related("category").prefetch_related(
                    Prefetch(
                        lookup="fittings",
                        queryset=Fitting.objects.select_related("ship_type"),
                    )
                )
            )
            .order_by("name")
        )
    else:
        # Alliance Auth AFAT
        from afat.models import Doctrine  # pylint: disable=import-outside-toplevel

        doctrines = Doctrine.objects.filter(is_enabled=True).distinct().order_by("name")

    return doctrines



"""
Helper functions for association between Auth user and EveCharacter
"""

# Django
from django.contrib.auth.models import User
from django.db.models import QuerySet

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter


def get_all_characters_with_fats_from_user(user: User) -> QuerySet[EveCharacter]:
    """
    Get all characters from a user

    :param user: The user
    :type user: User
    :return: A queryset of EveCharacter objects
    :rtype: QuerySet[EveCharacter]
    """

    return (
        EveCharacter.objects.select_related("character_ownership")
        .filter(character_ownership__user=user, afat_fats__isnull=False)
        .order_by("-userprofile", "character_name")
        .distinct()
    )
